@@ -1,7 +1,7 @@
 // SPDX-FileCopyrightText: 2021 Heiko Schaefer <heiko@schaefer.name>
 // SPDX-License-Identifier: MIT OR Apache-2.0
 
-use anyhow::Result;
+use anyhow::{anyhow, Result};
 use std::path::{Path, PathBuf};
 use structopt::StructOpt;
 
@@ -34,28 +34,26 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
         cli::Command::Status { ident, verbose } => {
             print_status(ident, verbose)?;
         }
+        cli::Command::SetIdentity { ident, id } => {
+            set_identity(&ident, id)?;
+        }
         cli::Command::Decrypt {
             ident,
-            pin_file,
+            user_pin,
             cert_file,
             input,
         } => {
-            decrypt(&ident, &pin_file, &cert_file, input.as_deref())?;
+            decrypt(&ident, user_pin, &cert_file, input.as_deref())?;
         }
         cli::Command::Sign {
             ident,
-            pin_file,
+            user_pin,
             cert_file,
             detached,
             input,
         } => {
             if detached {
-                sign_detached(
-                    &ident,
-                    &pin_file,
-                    &cert_file,
-                    input.as_deref(),
-                )?;
+                sign_detached(&ident, user_pin, &cert_file, input.as_deref())?;
             } else {
                 return Err(anyhow::anyhow!(
                     "Only detached signatures are supported for now"
@@ -68,20 +66,22 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
         }
         cli::Command::Admin {
             ident,
-            pin_file,
+            admin_pin,
             cmd,
         } => {
-            let mut card = util::open_card(&ident)?.into();
-            let mut open = Open::open(&mut card)?;
+            let mut card = util::open_card(&ident)?;
+            let mut open = Open::new(&mut card)?;
 
             match cmd {
                 cli::AdminCommand::Name { name } => {
-                    let mut admin = util::get_admin(&mut open, &pin_file)?;
+                    let mut admin =
+                        util::verify_to_admin(&mut open, admin_pin)?;
 
                     let _ = admin.set_name(&name)?;
                 }
                 cli::AdminCommand::Url { url } => {
-                    let mut admin = util::get_admin(&mut open, &pin_file)?;
+                    let mut admin =
+                        util::verify_to_admin(&mut open, admin_pin)?;
 
                     let _ = admin.set_url(&url)?;
                 }
@@ -91,8 +91,8 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
                     dec_fp,
                     auth_fp,
                 } => {
-                    let admin = util::get_admin(&mut open, &pin_file)?;
                     let key = Cert::from_file(keyfile)?;
+                    let admin = util::verify_to_admin(&mut open, admin_pin)?;
 
                     if (&sig_fp, &dec_fp, &auth_fp) == (&None, &None, &None) {
                         // If no fingerprint has been provided, we check if
@@ -106,19 +106,16 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
                     }
                 }
                 cli::AdminCommand::Generate {
-                    user_pin_file,
+                    user_pin,
                     output,
                     no_decrypt,
                     no_auth,
                     algo,
                 } => {
-                    let pw3 = util::get_pin(&pin_file)?;
-                    let pw1 = util::get_pin(&user_pin_file)?;
-
                     generate_keys(
                         open,
-                        &pw3,
-                        &pw1,
+                        admin_pin,
+                        user_pin,
                         output,
                         !no_decrypt,
                         !no_auth,
@@ -137,9 +134,8 @@ fn list_cards() -> Result<()> {
     if !cards.is_empty() {
         println!("Available OpenPGP cards:");
 
-        for card in cards {
-            let mut card = card.into();
-            let open = Open::open(&mut card)?;
+        for mut card in cards {
+            let open = Open::new(&mut card)?;
             println!(" {}", open.application_identifier()?.ident());
         }
     } else {
@@ -148,19 +144,29 @@ fn list_cards() -> Result<()> {
     Ok(())
 }
 
+fn set_identity(
+    ident: &str,
+    id: u8,
+) -> Result<(), Box<dyn std::error::Error>> {
+    let mut card = util::open_card(ident)?;
+
+    card.set_identity(id)?;
+
+    Ok(())
+}
+
 fn print_status(ident: Option<String>, verbose: bool) -> Result<()> {
-    let ccb = if let Some(ident) = ident {
+    let mut ca = if let Some(ident) = ident {
         util::open_card(&ident)?
     } else {
         let mut cards = util::cards()?;
         if cards.len() == 1 {
             cards.pop().unwrap()
         } else {
-            return Err(anyhow::anyhow!("Found {} cards", cards.len()).into());
+            return Err(anyhow::anyhow!("Found {} cards", cards.len()));
         }
     };
-    let mut card = ccb.into();
-    let mut open = Open::open(&mut card)?;
+    let mut open = Open::new(&mut ca)?;
 
     print!("OpenPGP card {}", open.application_identifier()?.ident());
 
@@ -210,39 +216,50 @@ fn print_status(ident: Option<String>, verbose: bool) -> Result<()> {
     let kgt = open.key_generation_times()?;
 
     println!();
-    println!(
-        "Signature key ({})",
-        open.algorithm_attributes(KeyType::Signing)?,
-    );
+    println!("Signature key");
     if let Some(fp) = fps.signature() {
         println!("  fingerprint: {}", fp.to_spaced_hex());
     }
     if let Some(kgt) = kgt.signature() {
-        println! {"  created: {}",kgt.formatted()};
+        println! {"  created: {}", kgt.formatted()};
+    }
+    println! {"  algorithm: {}", open.algorithm_attributes(KeyType::Signing)?};
+
+    if verbose {
+        if let Ok(pkm) = open.public_key(KeyType::Signing) {
+            println! {"  public key material: {}", pkm};
+        }
     }
 
     println!();
-    println!(
-        "Decryption key ({})",
-        open.algorithm_attributes(KeyType::Decryption)?,
-    );
+    println!("Decryption key");
     if let Some(fp) = fps.decryption() {
         println!("  fingerprint: {}", fp.to_spaced_hex());
     }
     if let Some(kgt) = kgt.decryption() {
-        println! {"  created: {}",kgt.formatted()};
+        println! {"  created: {}", kgt.formatted()};
+    }
+    println! {"  algorithm: {}", open.algorithm_attributes(KeyType::Decryption)?};
+
+    if verbose {
+        if let Ok(pkm) = open.public_key(KeyType::Decryption) {
+            println! {"  public key material: {}", pkm};
+        }
     }
 
     println!();
-    println!(
-        "Authentication key ({})",
-        open.algorithm_attributes(KeyType::Authentication)?,
-    );
+    println!("Authentication key");
     if let Some(fp) = fps.authentication() {
         println!("  fingerprint: {}", fp.to_spaced_hex());
     }
     if let Some(kgt) = kgt.authentication() {
-        println! {"  created: {}",kgt.formatted()};
+        println! {"  created: {}", kgt.formatted()};
+    }
+    println! {"  algorithm: {}", open.algorithm_attributes(KeyType::Authentication)?};
+    if verbose {
+        if let Ok(pkm) = open.public_key(KeyType::Authentication) {
+            println! {"  public key material: {}", pkm};
+        }
     }
 
     // technical details about the card and its state
@@ -250,31 +267,40 @@ fn print_status(ident: Option<String>, verbose: bool) -> Result<()> {
     println!();
 
     let sst = open.security_support_template()?;
-    println!("Signature counter: {}", sst.get_signature_count());
+    println!("Signature counter: {}", sst.signature_count());
 
     let pws = open.pw_status_bytes()?;
 
     println!(
         "Signature pin only valid once: {}",
-        pws.get_pw1_cds_valid_once()
+        pws.pw1_cds_valid_once()
     );
 
     println!("Password validation retry count:");
     println!(
         "  user pw: {}, reset: {}, admin pw: {}",
-        pws.get_err_count_pw1(),
-        pws.get_err_count_rst(),
-        pws.get_err_count_pw3(),
+        pws.err_count_pw1(),
+        pws.err_count_rc(),
+        pws.err_count_pw3(),
     );
 
     // FIXME: add General key info; login data; KDF setting
 
     if verbose {
-        if let Some(ai) = open.algorithm_information()? {
+        // Algorithm information (list of supported algorithms)
+        if let Ok(Some(ai)) = open.algorithm_information() {
             println!();
             println!("Supported algorithms:");
             println!("{}", ai);
         }
+
+        // YubiKey specific (?) firmware version
+        if let Ok(ver) = open.firmware_version() {
+            let ver =
+                ver.iter().map(u8::to_string).collect::<Vec<_>>().join(".");
+
+            println!("Firmware Version: {}", ver);
+        }
     }
 
     Ok(())
@@ -282,7 +308,7 @@ fn print_status(ident: Option<String>, verbose: bool) -> Result<()> {
 
 fn decrypt(
     ident: &str,
-    pin_file: &Path,
+    pin_file: Option<PathBuf>,
     cert_file: &Path,
     input: Option<&Path>,
 ) -> Result<(), Box<dyn std::error::Error>> {
@@ -291,10 +317,10 @@ fn decrypt(
 
     let input = util::open_or_stdin(input.as_deref())?;
 
-    let mut card = util::open_card(&ident)?.into();
-    let mut open = Open::open(&mut card)?;
+    let mut card = util::open_card(ident)?;
+    let mut open = Open::new(&mut card)?;
 
-    let mut user = util::get_user(&mut open, &pin_file)?;
+    let mut user = util::verify_to_user(&mut open, pin_file)?;
     let d = user.decryptor(&cert, &p)?;
 
     let db = DecryptorBuilder::from_reader(input)?;
@@ -307,7 +333,7 @@ fn decrypt(
 
 fn sign_detached(
     ident: &str,
-    pin_file: &Path,
+    pin_file: Option<PathBuf>,
     cert_file: &Path,
     input: Option<&Path>,
 ) -> Result<(), Box<dyn std::error::Error>> {
@@ -316,10 +342,10 @@ fn sign_detached(
 
     let mut input = util::open_or_stdin(input.as_deref())?;
 
-    let mut card = util::open_card(&ident)?.into();
-    let mut open = Open::open(&mut card)?;
+    let mut card = util::open_card(ident)?;
+    let mut open = Open::new(&mut card)?;
 
-    let mut sign = util::get_sign(&mut open, &pin_file)?;
+    let mut sign = util::verify_to_sign(&mut open, pin_file)?;
     let s = sign.signer(&cert, &p)?;
 
     let message = Armorer::new(Message::new(std::io::stdout())).build()?;
@@ -333,27 +359,18 @@ fn sign_detached(
 
 fn factory_reset(ident: &str) -> Result<()> {
     println!("Resetting Card {}", ident);
-    let mut card = util::open_card(ident)?.into();
-    Open::open(&mut card)?.factory_reset()
+    let mut card = util::open_card(ident)?;
+    Open::new(&mut card)?.factory_reset()
 }
 
 fn key_import_yolo(mut admin: Admin, key: &Cert) -> Result<()> {
     let p = StandardPolicy::new();
 
-    let sig =
-        openpgp_card_sequoia::sq_util::get_subkey(&key, &p, KeyType::Signing)?;
+    let sig = sq_util::subkey_by_type(key, &p, KeyType::Signing)?;
 
-    let dec = openpgp_card_sequoia::sq_util::get_subkey(
-        &key,
-        &p,
-        KeyType::Decryption,
-    )?;
+    let dec = sq_util::subkey_by_type(key, &p, KeyType::Decryption)?;
 
-    let auth = openpgp_card_sequoia::sq_util::get_subkey(
-        &key,
-        &p,
-        KeyType::Authentication,
-    )?;
+    let auth = sq_util::subkey_by_type(key, &p, KeyType::Authentication)?;
 
     if let Some(sig) = sig {
         println!("Uploading {} as signing key", sig.fingerprint());
@@ -382,7 +399,7 @@ fn key_import_explicit(
 
     if let Some(sig_fp) = sig_fp {
         if let Some(sig) =
-            sq_util::get_subkey_by_fingerprint(&key, &p, &sig_fp)?
+            sq_util::private_subkey_by_fingerprint(key, &p, &sig_fp)?
         {
             println!("Uploading {} as signing key", sig.fingerprint());
             admin.upload_key(sig, KeyType::Signing, None)?;
@@ -393,7 +410,7 @@ fn key_import_explicit(
 
     if let Some(dec_fp) = dec_fp {
         if let Some(dec) =
-            sq_util::get_subkey_by_fingerprint(&key, &p, &dec_fp)?
+            sq_util::private_subkey_by_fingerprint(key, &p, &dec_fp)?
         {
             println!("Uploading {} as decryption key", dec.fingerprint());
             admin.upload_key(dec, KeyType::Decryption, None)?;
@@ -404,7 +421,7 @@ fn key_import_explicit(
 
     if let Some(auth_fp) = auth_fp {
         if let Some(auth) =
-            sq_util::get_subkey_by_fingerprint(&key, &p, &auth_fp)?
+            sq_util::private_subkey_by_fingerprint(key, &p, &auth_fp)?
         {
             println!("Uploading {} as authentication key", auth.fingerprint());
             admin.upload_key(auth, KeyType::Authentication, None)?;
@@ -418,8 +435,8 @@ fn key_import_explicit(
 
 fn generate_keys(
     mut open: Open,
-    pw3: &str,
-    pw1: &str,
+    pw3_path: Option<PathBuf>,
+    pw1_path: Option<PathBuf>,
     output: Option<PathBuf>,
     decrypt: bool,
     auth: bool,
@@ -438,47 +455,58 @@ fn generate_keys(
     // Because of this, for generation of RSA keys, here we take the approach
     // of first trying one variant, and then if that fails, try the other.
 
-    let algos = match algo.as_deref() {
-        None => vec![],
-        Some("rsa2048") => vec![AlgoSimple::RSA2k(32), AlgoSimple::RSA2k(17)],
-        Some("rsa3072") => vec![AlgoSimple::RSA3k(32), AlgoSimple::RSA3k(17)],
-        Some("rsa4096") => vec![AlgoSimple::RSA4k(32), AlgoSimple::RSA4k(17)],
-        Some("nistp256") => vec![AlgoSimple::NIST256],
-        Some("nistp384") => vec![AlgoSimple::NIST384],
-        Some("nistp521") => vec![AlgoSimple::NIST521],
-        Some("25519") => vec![AlgoSimple::Curve25519],
-        _ => unimplemented!("unexpected algorithm"),
+    let a = match algo.as_deref() {
+        None => None,
+        Some("rsa2048") => Some(AlgoSimple::RSA2k),
+        Some("rsa3072") => Some(AlgoSimple::RSA3k),
+        Some("rsa4096") => Some(AlgoSimple::RSA4k),
+        Some("nistp256") => Some(AlgoSimple::NIST256),
+        Some("nistp384") => Some(AlgoSimple::NIST384),
+        Some("nistp521") => Some(AlgoSimple::NIST521),
+        Some("25519") => Some(AlgoSimple::Curve25519),
+        _ => return Err(anyhow!("Unexpected algorithm")),
     };
 
-    log::info!(
-        " Key generation will be attempted with these algos: {:?}",
-        algos
-    );
+    log::info!(" Key generation will be attempted with algo: {:?}", a);
 
     // 2) Then, generate keys on the card.
     // We need "admin" access to the card for this).
-
-    open.verify_admin(pw3)?;
-
     let (key_sig, key_dec, key_aut) = {
-        if let Some(mut admin) = open.admin_card() {
-            gen_subkeys(&mut admin, decrypt, auth, algos)?
+        if let Ok(mut admin) = util::verify_to_admin(&mut open, pw3_path) {
+            gen_subkeys(&mut admin, decrypt, auth, a)?
         } else {
-            // FIXME: couldn't get admin mode
-            unimplemented!()
+            return Err(anyhow!("Failed to open card in admin mode."));
         }
     };
 
     // 3) Generate a Cert from the generated keys. For this, we
     // need "signing" access to the card (to make binding signatures within
     // the Cert).
+    let pin = if let Some(pw1) = pw1_path {
+        Some(util::load_pin(&pw1)?)
+    } else {
+        if open.feature_pinpad_verify() {
+            println!();
+            println!(
+                "Next: generating your public cert. You will need to enter \
+                your user PIN multiple times to make binding signatures."
+            );
+        } else {
+            return Err(anyhow!(
+                "No user PIN file provided, and no pinpad found"
+            ));
+        }
+        None
+    };
 
-    let cert = make_cert(&mut open, key_sig, key_dec, key_aut, pw1)?;
+    let cert = make_cert(&mut open, key_sig, key_dec, key_aut, pin, &|| {
+        println!("Enter user PIN on card reader pinpad.")
+    })?;
     let armored = String::from_utf8(cert.armored().to_vec()?)?;
 
     // Write armored certificate to the output file (or stdout)
     let mut output = util::open_or_stdout(output.as_deref())?;
-    output.write(armored.as_bytes())?;
+    output.write_all(armored.as_bytes())?;
 
     Ok(())
 }
@@ -487,91 +515,19 @@ fn gen_subkeys(
     admin: &mut Admin,
     decrypt: bool,
     auth: bool,
-    algos: Vec<AlgoSimple>,
+    algo: Option<AlgoSimple>,
 ) -> Result<(PublicKey, Option<PublicKey>, Option<PublicKey>)> {
-    // We begin with the signing subkey, which is mandatory.
-    // If there are multiple algorithms we should try (i.e. two rsa
-    // variants), we'll attempt to make the signing subkey with each of
-    // them, and continue with the first one that works, if any.
-
+    // We begin by generating the signing subkey, which is mandatory.
     println!(" Generate subkey for Signing");
-    let (alg, key_sig) = if algos.is_empty() {
-        // Handle unset algorithm from user (-> leave algo as is on the card)
-        log::info!(" Running key generation with implicit algo");
-
-        let (pkm, ts) = admin.generate_key_simple(KeyType::Signing, None)?;
-        let key = public_key_material_to_key(&pkm, KeyType::Signing, ts)?;
-
-        (None, key)
-    } else {
-        // Try list of algos until one works - or return list of all errors.
-
-        let mut errors = vec![];
-
-        let mut algo: Option<AlgoSimple> = None;
-        let mut key_sig: Option<PublicKey> = None;
-
-        // Check each algo while making key_sig.
-        //
-        // Return the result for the first one that works, and keep using
-        // that algo.
-        //
-        // If none of them work, fail and return all errors
-
-        for (n, alg) in algos.iter().enumerate() {
-            let a = Some(alg.clone());
-
-            log::info!(" Trying key generation with algo {:?}", alg);
-
-            match admin.generate_key_simple(KeyType::Signing, a) {
-                Ok((pkm, ts)) => {
-                    // generated a valid key -> return it
-                    let key = public_key_material_to_key(
-                        &pkm,
-                        KeyType::Signing,
-                        ts,
-                    )?;
-                    algo = a;
-                    key_sig = Some(key);
-                    break;
-                }
-                Err(e) => match (n, n == algos.len() - 1) {
-                    // there was only one algo, and it failed
-                    (0, true) => return Err(e.into()),
-
-                    // there was an error, but there are more algo to try
-                    (_, false) => errors.push(e),
-
-                    // there were multiple algo, there was an error, and
-                    // this is the last algo
-                    (_, true) => {
-                        errors.push(e);
-
-                        let err = anyhow::anyhow!(
-                            "Key generation failed for all algorithm \
-                            variants {:x?} ({:?})",
-                            errors,
-                            algos
-                        );
-                        return Err(err);
-                    }
-                },
-            };
-        }
-
-        // Neither of these should be possible, but the compiler can't tell.
-        assert!(algo.is_some());
-        assert!(key_sig.is_some());
-
-        (algo, key_sig.unwrap())
-    };
+    let (pkm, ts) = admin.generate_key_simple(KeyType::Signing, algo)?;
+    let key_sig = public_key_material_to_key(&pkm, KeyType::Decryption, ts)?;
 
     // make decryption subkey (unless disabled), with the same algorithm as
     // the sig key
-
     let key_dec = if decrypt {
         println!(" Generate subkey for Decryption");
-        let (pkm, ts) = admin.generate_key_simple(KeyType::Decryption, alg)?;
+        let (pkm, ts) =
+            admin.generate_key_simple(KeyType::Decryption, algo)?;
         Some(public_key_material_to_key(&pkm, KeyType::Decryption, ts)?)
     } else {
         None
@@ -579,11 +535,10 @@ fn gen_subkeys(
 
     // make authentication subkey (unless disabled), with the same
     // algorithm as the sig key
-
     let key_aut = if auth {
         println!(" Generate subkey for Authentication");
         let (pkm, ts) =
-            admin.generate_key_simple(KeyType::Authentication, alg)?;
+            admin.generate_key_simple(KeyType::Authentication, algo)?;
 
         Some(public_key_material_to_key(
             &pkm,
